@@ -22,7 +22,7 @@ import pickle as pkl
 from joblib import Parallel, delayed, wrap_non_picklable_objects
 
 
-def generate_coordinates(infile, outfile=None, run_id=None, target_peptides=None, m_score=0.05, use_transition_peptide_mapping=False, use_only_detecting_transitions=True, verbose=0):
+def generate_coordinates(infile, outfile=None, run_id=None, target_peptides=None, m_score=0.05, use_transition_peptide_mapping=False, use_only_detecting_transitions=True, verbose=0, log_file=None):
     """
     Generate a dictionary of target coordinates to extract from Raw diaPASEF mzML data
 
@@ -33,6 +33,10 @@ def generate_coordinates(infile, outfile=None, run_id=None, target_peptides=None
       a pickle file containing a dictionary of target coordinates
     """
 
+    # Initialise logger
+    if log_file is not None:
+        setup_logger(log_file, verbose)
+
     if infile.lower().endswith("osw"):
         if verbose == 10:
             logging.debug(
@@ -57,6 +61,22 @@ def generate_coordinates(infile, outfile=None, run_id=None, target_peptides=None
             # Restrict to top peak group rank to generate coordinates for. We also filter based on MS2 QVALUE to reduce the number of peptides to generate coordinates for
             join_on_score_table = "INNER JOIN (SELECT * FROM SCORE_MS2 WHERE RANK ==1 AND QVALUE < %s) AS SCORE_TABLE ON SCORE_TABLE.FEATURE_ID = FEATURE.ID" % (m_score)
 
+        # MS1 Feature Identificaion
+        if check_sqlite_table(con, "FEATURE_MS1"):
+            join_on_feature_ms1_table = "INNER JOIN FEATURE_MS1 ON FEATURE.ID = FEATURE_MS1.FEATURE_ID"
+            select_feature_ms1_stmt = "FEATURE_MS1.AREA_INTENSITY AS ms1_intensity,"
+        else:
+            join_on_feature_ms1_table = ""
+            select_feature_ms1_stmt = ""
+
+        # MS2 Feature Identificaion
+        if check_sqlite_table(con, "FEATURE_MS2"):
+            join_on_feature_ms2_table = "INNER JOIN FEATURE_MS2 ON FEATURE.ID = FEATURE_MS2.FEATURE_ID"
+            select_feature_ms2_stmt = "FEATURE_MS2.AREA_INTENSITY AS intensity,"
+        else:
+            join_on_feature_ms2_table = ""
+            select_feature_ms2_stmt = ""
+
         # Get Run ID information
         run_ids_df = pd.read_sql_query("SELECT * FROM RUN", con)
         if run_id is None:
@@ -87,7 +107,7 @@ def generate_coordinates(infile, outfile=None, run_id=None, target_peptides=None
             logging.info(
                 f"Generating coordinates for the following peptides: {target_peptides}")
 
-            sql_query = '''
+            sql_query = f'''
               SELECT
               PEPTIDE.MODIFIED_SEQUENCE AS peptide,
               PRECURSOR.PRECURSOR_MZ AS precursor_mz,
@@ -100,29 +120,34 @@ def generate_coordinates(infile, outfile=None, run_id=None, target_peptides=None
               FEATURE.LEFT_WIDTH AS left_width,
               FEATURE.RIGHT_WIDTH AS right_width,
               FEATURE.EXP_IM AS im_apex,
-              SCORE_TABLE.QVALUE as qvalue
+              {select_feature_ms1_stmt}
+              {select_feature_ms2_stmt}
+              SCORE_TABLE.QVALUE as qvalue,
+              PRECURSOR.DECOY as decoy
               FROM PRECURSOR
               INNER JOIN PRECURSOR_PEPTIDE_MAPPING ON PRECURSOR_PEPTIDE_MAPPING.PRECURSOR_ID = PRECURSOR.ID
               INNER JOIN (
                 SELECT * 
                 FROM PEPTIDE
-                WHERE PEPTIDE.MODIFIED_SEQUENCE IN ("%s")
+                WHERE PEPTIDE.MODIFIED_SEQUENCE IN ("{'","'.join(target_peptides)}")
                 ) AS PEPTIDE ON PEPTIDE.ID = PRECURSOR_PEPTIDE_MAPPING.PEPTIDE_ID
-              %s 
+              {use_transition_map} 
               INNER JOIN (
                 SELECT *
                 FROM TRANSITION
-                WHERE TRANSITION.DETECTING in (%s)
-              ) AS TRANSITION ON TRANSITION.ID = %s
+                WHERE TRANSITION.DETECTING in ({detecting_col})
+              ) AS TRANSITION ON TRANSITION.ID = {join_on_transition_table_id}
               INNER JOIN (
                 SELECT *
                 FROM FEATURE
-                WHERE FEATURE.RUN_ID=%s
+                WHERE FEATURE.RUN_ID={run_id}
               ) AS FEATURE ON FEATURE.PRECURSOR_ID = PRECURSOR.ID
-              %s
-              ''' % ('","'.join(target_peptides), use_transition_map, detecting_col, join_on_transition_table_id, run_id, join_on_score_table )
+              {join_on_feature_ms1_table}
+              {join_on_feature_ms2_table}
+              {join_on_score_table}
+              ''' 
         else: # if targeted peptides is "None" than select all peptides
-            sql_query = '''
+            sql_query = f'''
               SELECT
               PEPTIDE.MODIFIED_SEQUENCE AS peptide,
               PRECURSOR.PRECURSOR_MZ AS precursor_mz,
@@ -135,23 +160,28 @@ def generate_coordinates(infile, outfile=None, run_id=None, target_peptides=None
               FEATURE.LEFT_WIDTH AS left_width,
               FEATURE.RIGHT_WIDTH AS right_width,
               FEATURE.EXP_IM AS im_apex,
-              SCORE_TABLE.QVALUE as qvalue
+              {select_feature_ms1_stmt}
+              {select_feature_ms2_stmt}
+              SCORE_TABLE.QVALUE as qvalue,
+              PRECURSOR.DECOY as decoy
               FROM PRECURSOR
               INNER JOIN PRECURSOR_PEPTIDE_MAPPING ON PRECURSOR_PEPTIDE_MAPPING.PRECURSOR_ID = PRECURSOR.ID
               INNER JOIN PEPTIDE ON PEPTIDE.ID = PRECURSOR_PEPTIDE_MAPPING.PEPTIDE_ID
-              %s
+              {use_transition_map}
               INNER JOIN (
                 SELECT *
                 FROM TRANSITION
-                WHERE TRANSITION.DETECTING in (%s)
-              ) AS TRANSITION ON TRANSITION.ID = %s
+                WHERE TRANSITION.DETECTING in ({detecting_col})
+              ) AS TRANSITION ON TRANSITION.ID = {join_on_transition_table_id}
               INNER JOIN (
                 SELECT *
                 FROM FEATURE
-                WHERE FEATURE.RUN_ID=%s
+                WHERE FEATURE.RUN_ID={run_id}
               ) AS FEATURE ON FEATURE.PRECURSOR_ID = PRECURSOR.ID
-              %s
-              ''' % (use_transition_map, detecting_col, join_on_transition_table_id, run_id, join_on_score_table )
+              {join_on_feature_ms1_table}
+              {join_on_feature_ms2_table}
+              {join_on_score_table}
+              '''
 
         if verbose == 10:
             logging.debug(f"Injecting SQL Query:\n{sql_query}")
@@ -260,7 +290,7 @@ class data_io():
             self.filtered = exp
         else:
             if self.readOptions=="ondisk":
-                # TODO: For some reason the OnDiscExperiment doesn't read StringMetaData, which is currecntly used to store filtered IM data
+                # TODO: For some reason the OnDiscExperiment doesn't read StringMetaData, which is currecntly used to store filtered IM data. This should no longer be an issue..
                 with code_block_timer('Creating OnDiscExperiment...', logging.debug):
                     exp = po.OnDiscMSExperiment()
 
@@ -378,7 +408,19 @@ class data_io():
                     logging.debug(
                         f"Adding MS{spec.getMSLevel()} spectrum for peptide: {spec.getMetaValue('peptide')} with native id: {spec.getNativeID()}")
                 add_df = pd.DataFrame({'native_id': spec.getNativeID(), 'ms_level': spec.getMSLevel(), 'peptide': spec.getMetaValue(
-                    'peptide'), 'precursor_mz': precursor.getMZ(), 'charge': precursor.getCharge(), 'mz': mz, 'rt': rt, 'im': im, 'int': intensity})
+                    'peptide'), 'precursor_mz': spec.getMetaValue('precursor_mz'), 'charge': precursor.getCharge(), 'mz': mz, 'rt': rt, 'im': im, 'int': intensity})
+                ## Add additional meta data
+                meta_data = pd.DataFrame()
+                if spec.getMetaValue('rt_apex') is not None:
+                    meta_data['rt_apex'] = pd.Series(spec.getMetaValue('rt_apex'))
+                if spec.getMetaValue('im_apex') is not None:
+                    meta_data['im_apex'] = pd.Series(spec.getMetaValue('im_apex'))
+                if spec.getMetaValue('rt_left_width') is not None:
+                    meta_data['rt_left_width'] = pd.Series(spec.getMetaValue('rt_left_width'))    
+                if spec.getMetaValue('rt_right_width') is not None:
+                    meta_data['rt_right_width'] = pd.Series(spec.getMetaValue('rt_right_width'))
+                if not meta_data.empty:
+                    add_df = add_df.join(meta_data, how='cross')
                 results_df = pd.concat([results_df, add_df])
         logging.info(
             f"Saving filtered target spectra data to tabular tsv format: {out_file}")
@@ -503,7 +545,7 @@ class TargeteddiaPASEFExperiment(data_io):
         return any([check_mz >= bounds[0] and check_mz <= bounds[1] for bounds in target_product_upper_lower_list])
 
     # @wrap_non_picklable_objects # seems to cause an error with spectrum. list not having attribute getRT()
-    def filter_single_spectrum(self, spec_indice, mslevel, target_peptide_group, rt_start, rt_end, im_start, im_end, target_precursor_mz_lower, target_precursor_mz_upper, target_product_upper_lower_list, verbose=0):
+    def filter_single_spectrum(self, spec_indice, mslevel, target_peptide_group, rt_start, rt_end, im_start, im_end, target_precursor_mz, target_precursor_mz_lower, target_precursor_mz_upper, target_product_upper_lower_list, verbose=0):
         """
         Filter a single spectrum for a given spectrum indice.
 
@@ -541,6 +583,13 @@ class TargeteddiaPASEFExperiment(data_io):
         else:
             click.ClickException(f"ERROR: Unknown readOptions ({self.readOptions}) given! Has to be one of 'ondisk', 'cached'")
         
+        # Get SWATH windows upper and lower
+        dir(spec)
+        current_prec = spec.getPrecursors()[0]
+        dir(current_prec)
+        swath_mz_win_lower = current_prec.getMZ() - current_prec.getIsolationWindowLowerOffset()
+        swath_mz_win_upper = current_prec.getMZ() + current_prec.getIsolationWindowUpperOffset()
+
         if (self.readOptions=="ondisk"): # TODO: Can we get rid of these annoying if statement checks? I guess we do this because in ondisk exp we need the get_data view
             im_match_bool = (im_array[0].get_data() > im_start) & (
                 im_array[0].get_data() < im_end)
@@ -555,11 +604,18 @@ class TargeteddiaPASEFExperiment(data_io):
                 logging.debug(
                     f"Adding MS1 spectrum {spec.getNativeID()} with spectrum indice {spec_indice} filtered for {sum(mz_match_bool*im_match_bool)} spectra between {target_precursor_mz_lower} m/z and {target_precursor_mz_upper} m/z and IM between {im_start} and {im_end}")
         elif spec.getMSLevel() == 2 and 2 in mslevel:
-            mz_match_bool = np.array(list(map(self.is_mz_in_product_mz_tol_window, mz_array, itertools.repeat(
-                target_product_upper_lower_list, len(mz_array)))))
-            if verbose == 10 and any(mz_match_bool*im_match_bool):
-                logging.debug(
-                    f"Adding MS2 spectrum {spec.getNativeID()} with spectrum indice {spec_indice} filtered for {sum(mz_match_bool*im_match_bool)} spectra between {target_product_upper_lower_list} m/z and IM between {im_start} and {im_end}")
+            # Only extract product spectra if current spectrums isolation window contains precursor mz
+            if target_precursor_mz > swath_mz_win_lower and target_precursor_mz < swath_mz_win_upper:
+                mz_match_bool = np.array(list(map(self.is_mz_in_product_mz_tol_window, mz_array, itertools.repeat(
+                    target_product_upper_lower_list, len(mz_array)))))
+                if verbose == 10 and any(mz_match_bool*im_match_bool):
+                    logging.debug(
+                        f"Adding MS2 spectrum {spec.getNativeID()} with spectrum indice {spec_indice} filtered for {sum(mz_match_bool*im_match_bool)} spectra between {target_product_upper_lower_list} m/z and IM between {im_start} and {im_end}")
+            else:
+                mz_match_bool = np.zeros(mz_array.shape).astype(bool)
+                if verbose == 10:
+                    logging.debug(
+                        f"Skipping MS2 spectrum {spec.getNativeID()} with spectrum indice {spec_indice} because current swath isolation window ({swath_mz_win_lower} m/z - {swath_mz_win_upper} m/z) does not contain target precursor m/z ({target_precursor_mz})")
         # Only write out filtered spectra if there is any fitlered spectra to write out
         if any(mz_match_bool*im_match_bool):
             with code_block_timer(f'Getting filtered spectrum...', logging.debug):
@@ -582,17 +638,34 @@ class TargeteddiaPASEFExperiment(data_io):
                 # Set peptide meta data
                 spec.setMetaValue(
                     'peptide', self.peptides[target_peptide_group]['peptide'])
+                spec.setMetaValue(
+                    'precursor_mz', self.peptides[target_peptide_group]['precursor_mz'])    
+                [spec.setMetaValue(f'product_mz_{i}', mz) for i, mz in zip(range(0, len(self.peptides[target_peptide_group]['product_mz'])), self.peptides[target_peptide_group]['product_mz'])]
+                if 'rt_apex' in self.peptides[target_peptide_group].keys():
+                    spec.setMetaValue(
+                    'rt_apex', self.peptides[target_peptide_group]['rt_apex'])
+                if 'im_apex' in self.peptides[target_peptide_group].keys():
+                    spec.setMetaValue(
+                    'im_apex', self.peptides[target_peptide_group]['im_apex'])
+                if 'rt_boundaries' in self.peptides[target_peptide_group].keys():
+                    spec.setMetaValue(
+                    'rt_left_width', self.peptides[target_peptide_group]['rt_boundaries'][0])
+                    spec.setMetaValue(
+                    'rt_right_width', self.peptides[target_peptide_group]['rt_boundaries'][1])
                 precursor = po.Precursor()
                 precursor.setCharge(
                     self.peptides[target_peptide_group]['charge'])
                 precursor.setMZ(
-                    self.peptides[target_peptide_group]['precursor_mz'])
+                    current_prec.getMZ())
+                precursor.setIsolationWindowLowerOffset(current_prec.getIsolationWindowLowerOffset())
+                precursor.setIsolationWindowUpperOffset(current_prec.getIsolationWindowUpperOffset())
                 spec.setPrecursors([precursor])
-
+                
                 # Set Prodcut mz values
-                spec.setProducts([self.set_product(
-                    mz, ) for mz in self.peptides[target_peptide_group]['product_mz']])
-
+                # Probably don't need to set products, since we use metaValue instead now.
+                # spec.setProducts([self.set_product(
+                    # mz, ) for mz in self.peptides[target_peptide_group]['product_mz']])
+            
             # Write out filtered spectra to file if consumer present. This is more memory efficient
             # TODO: Once paralization works, will need to think about how writing to consumer will be affected
             if self.consumer is not None:
@@ -682,7 +755,7 @@ class TargeteddiaPASEFExperiment(data_io):
                     #   File "stringsource", line 2, in pyopenms.pyopenms_1.MSSpectrum.__reduce_cython__
                     # TypeError: self.inst cannot be converted to a Python object for pickling
                     # TODO: Currently very memory heavy, maybe save filtered spectrums to disk? Done.
-                    filt_spec_list = Parallel(n_jobs=1)(delayed(self.filter_single_spectrum)(spectrum_indice, self.mslevel, target_peptide_group, rt_start, rt_end, im_start, im_end,
+                    filt_spec_list = Parallel(n_jobs=1)(delayed(self.filter_single_spectrum)(spectrum_indice, self.mslevel, target_peptide_group, rt_start, rt_end, im_start, im_end, target_precursor_mz,
                                                                                              target_precursor_mz_lower, target_precursor_mz_upper, target_product_upper_lower_list, self.verbose) for spectrum_indice in target_spectra_indices.tolist())
                 except:
                     traceback.print_exc(file=sys.stdout)
